@@ -29,7 +29,8 @@
 static int
 ReadStringFromToast(const char *buffer,
 		unsigned int buff_size,
-		unsigned int* out_size);
+		unsigned int* out_size,
+		int (*parse_value)(const char *, int));
 
 /*
  * Utilities for manipulation of header information for compressed
@@ -77,9 +78,15 @@ decode_timetz(const char *buffer, unsigned int buff_size, unsigned int *out_size
 static int
 decode_date(const char *buffer, unsigned int buff_size, unsigned int *out_size);
 
+static int
+decode_timestamp_internal(const char *buffer, unsigned int buff_size, unsigned int *out_size, bool with_timezone);
+
 static int
 decode_timestamp(const char *buffer, unsigned int buff_size, unsigned int *out_size);
 
+static int
+decode_timestamptz(const char *buffer, unsigned int buff_size, unsigned int *out_size);
+
 static int
 decode_float4(const char *buffer, unsigned int buff_size, unsigned int *out_size);
 
@@ -104,6 +111,12 @@ decode_char(const char *buffer, unsigned int buff_size, unsigned int *out_size);
 static int
 decode_name(const char *buffer, unsigned int buff_size, unsigned int *out_size);
 
+static int
+decode_numeric(const char *buffer, unsigned int buff_size, unsigned int *out_size);
+
+static int
+extract_data(const char *buffer, unsigned int buff_size, unsigned int *out_size, int (*parse_value)(const char *, int));
+
 static int
 decode_ignore(const char *buffer, unsigned int buff_size, unsigned int *out_size);
 
@@ -157,6 +170,9 @@ static ParseCallbackTableItem callback_table[] =
 	{
 		"timestamp", &decode_timestamp
 	},
+	{
+		"timestamptz", &decode_timestamptz
+	},
 	{
 		"real", &decode_float4
 	},
@@ -181,6 +197,9 @@ static ParseCallbackTableItem callback_table[] =
 	{
 		"name", &decode_name
 	},
+	{
+		"numeric", &decode_numeric
+	},
 	{
 		"char", &decode_char
 	},
@@ -263,7 +282,7 @@ CopyAppend(const char *str)
  * Append given string to current COPY line and encode special symbols
  * like \r, \n, \t and \\.
  */
-static void
+static int
 CopyAppendEncode(const char *str, int orig_len)
 {
 	/*
@@ -339,6 +358,7 @@ CopyAppendEncode(const char *str, int orig_len)
 
 	tmp_buff[curr_offset] = '\0';
 	CopyAppend(tmp_buff);
+	return 0;
 }
 
 /* CopyAppend version with format string support */
@@ -348,6 +368,144 @@ CopyAppendEncode(const char *str, int orig_len)
 	  CopyAppend(__copy_format_buff); \
   } while(0)
 
+/*
+ * Decode a numeric type and append the result to current COPY line
+ */
+static int
+CopyAppendNumeric(const char *buffer, int num_size)
+{
+       struct NumericData num;
+
+       num = *(struct NumericData *)buffer;
+       if (NUMERIC_IS_SPECIAL(&num))
+       {
+               if (NUMERIC_IS_NINF(&num))
+               {
+                       CopyAppend("-Infinity");
+                       return 0;
+               }
+               if (NUMERIC_IS_PINF(&num))
+               {
+                       CopyAppend("Infinity");
+                       return 0;
+               }
+               if (NUMERIC_IS_NAN(&num))
+               {
+                       CopyAppend("NaN");
+                       return 0;
+               }
+               return -2;
+       }
+       else
+       {
+               int				sign;
+               int				weight;
+               int				dscale;
+               int				ndigits;
+               int				i;
+               char			   *str;
+               char			   *cp;
+               char			   *endcp;
+               int				d;
+               bool				putit;
+               NumericDigit		d1;
+               NumericDigit		dig;
+               NumericDigit	   *digits;
+
+               sign = NUMERIC_SIGN(&num);
+               weight = NUMERIC_WEIGHT(&num);
+               dscale = NUMERIC_DSCALE(&num);
+
+               if (num_size == NUMERIC_HEADER_SIZE(&num))
+               {
+                       /* No digits - compressed zero. */
+                       CopyAppendFmt("%d", 0);
+                       return 0;
+               }
+               else
+               {
+                       ndigits = num_size / sizeof(NumericDigit);
+                       digits = (NumericDigit *)(buffer + NUMERIC_HEADER_SIZE(&num));
+                       i = (weight + 1) * DEC_DIGITS;
+                       if (i <= 0)
+                               i = 1;
+
+                       str = palloc(i + dscale + DEC_DIGITS + 2);
+                       cp = str;
+
+                       /*
+                        * Output a dash for negative values
+                        */
+                       if (sign == NUMERIC_NEG)
+                               *cp++ = '-';
+
+                       /*
+                        * Output all digits before the decimal point
+                        */
+                       if (weight < 0)
+                       {
+                               d = weight + 1;
+                               *cp++ = '0';
+                       }
+                       else
+                       {
+                               for (d = 0; d <= weight; d++)
+                               {
+                                       dig = (d < ndigits) ? digits[d] : 0;
+                                       /* In the first digit, suppress extra leading decimal zeroes */
+                                       putit = (d > 0);
+                                               d1 = dig / 1000;
+                                       dig -= d1 * 1000;
+                                       putit |= (d1 > 0);
+                                       if (putit)
+                                               *cp++ = d1 + '0';
+                                       d1 = dig / 100;
+                                       dig -= d1 * 100;
+                                       putit |= (d1 > 0);
+                                       if (putit)
+                                               *cp++ = d1 + '0';
+                                       d1 = dig / 10;
+                                       dig -= d1 * 10;
+                                       putit |= (d1 > 0);
+                                       if (putit)
+                                               *cp++ = d1 + '0';
+                                       *cp++ = dig + '0';
+                               }
+                       }
+
+                       /*
+                        * If requested, output a decimal point and all the digits that follow it.
+                        * We initially put out a multiple of DEC_DIGITS digits, then truncate if
+                        * needed.
+                        */
+                       if (dscale > 0)
+                       {
+                               *cp++ = '.';
+                               endcp = cp + dscale;
+                               for (i = 0; i < dscale; d++, i += DEC_DIGITS)
+                               {
+                                       dig = (d >= 0 && d < ndigits) ? digits[d] : 0;
+                                       d1 = dig / 1000;
+                                       dig -= d1 * 1000;
+                                       *cp++ = d1 + '0';
+                                       d1 = dig / 100;
+                                       dig -= d1 * 100;
+                                       *cp++ = d1 + '0';
+                                       d1 = dig / 10;
+                                       dig -= d1 * 10;
+                                       *cp++ = d1 + '0';
+                                       *cp++ = dig + '0';
+                               }
+                               cp = endcp;
+                       }
+                       *cp = '\0';
+                       CopyAppend(str);
+                       pfree(str);
+                       return 0;
+               }
+       }
+}
+
 /* Discard accumulated COPY line */
 static void
 CopyClear(void)
@@ -491,7 +649,7 @@ j2date(int jd, int *year, int *month, int *day)
 static int
 decode_smallint(const char *buffer, unsigned int buff_size, unsigned int *out_size)
 {
-	const char *new_buffer = (const char *) TYPEALIGN(sizeof(int16), (uintptr_t) buffer);
+	const char *new_buffer = (const char *) SHORTALIGN(buffer);
 	unsigned int delta = (unsigned int) ((uintptr_t) new_buffer - (uintptr_t) buffer);
 
 	if (buff_size < delta)
@@ -513,7 +671,7 @@ decode_smallint(const char *buffer, unsigned int buff_size, unsigned int *out_si
 static int
 decode_int(const char *buffer, unsigned int buff_size, unsigned int *out_size)
 {
-	const char *new_buffer = (const char *) TYPEALIGN(sizeof(int32), (uintptr_t) buffer);
+	const char *new_buffer = (const char *) INTALIGN(buffer);
 	unsigned int delta = (unsigned int) ((uintptr_t) new_buffer - (uintptr_t) buffer);
 
 	if (buff_size < delta)
@@ -534,7 +692,7 @@ decode_int(const char *buffer, unsigned int buff_size, unsigned int *out_size)
 static int
 decode_bigint(const char *buffer, unsigned int buff_size, unsigned int *out_size)
 {
-	const char *new_buffer = (const char *) TYPEALIGN(sizeof(int64), (uintptr_t) buffer);
+	const char *new_buffer = (const char *) LONGALIGN(buffer);
 	unsigned int delta = (unsigned int) ((uintptr_t) new_buffer - (uintptr_t) buffer);
 
 	if (buff_size < delta)
@@ -555,7 +713,7 @@ decode_bigint(const char *buffer, unsigned int buff_size, unsigned int *out_size
 static int
 decode_time(const char *buffer, unsigned int buff_size, unsigned int *out_size)
 {
-	const char *new_buffer = (const char *) TYPEALIGN(sizeof(int64), (uintptr_t) buffer);
+	const char *new_buffer = (const char *) LONGALIGN(buffer);
 	unsigned int delta = (unsigned int) ((uintptr_t) new_buffer - (uintptr_t) buffer);
 	int64		timestamp,
 				timestamp_sec;
@@ -584,7 +742,7 @@ decode_time(const char *buffer, unsigned int buff_size, unsigned int *out_size)
 static int
 decode_timetz(const char *buffer, unsigned int buff_size, unsigned int *out_size)
 {
-	const char *new_buffer = (const char *) TYPEALIGN(sizeof(int64), (uintptr_t) buffer);
+	const char *new_buffer = (const char *) LONGALIGN(buffer);
 	unsigned int delta = (unsigned int) ((uintptr_t) new_buffer - (uintptr_t) buffer);
 	int64		timestamp,
 				timestamp_sec;
@@ -617,9 +775,10 @@ decode_timetz(const char *buffer, unsigned int buff_size, unsigned int *out_size
 static int
 decode_date(const char *buffer, unsigned int buff_size, unsigned int *out_size)
 {
-	const char *new_buffer = (const char *) TYPEALIGN(sizeof(int32), (uintptr_t) buffer);
+	const char *new_buffer = (const char *) INTALIGN(buffer);
 	unsigned int delta = (unsigned int) ((uintptr_t) new_buffer - (uintptr_t) buffer);
-	int32		jd,
+	int32		d,
+				jd,
 				year,
 				month,
 				day;
@@ -635,7 +794,19 @@ decode_date(const char *buffer, unsigned int buff_size, unsigned int *out_size)
 
 	*out_size = sizeof(int32) + delta;
 
-	jd = *(int32 *) buffer + POSTGRES_EPOCH_JDATE;
+	d = *(int32 *) buffer;
+	if (d == PG_INT32_MIN)
+	{
+		CopyAppend("-infinity");
+		return 0;
+	}
+	if (d == PG_INT32_MAX)
+	{
+		CopyAppend("infinity");
+		return 0;
+	}
+
+	jd = d + POSTGRES_EPOCH_JDATE;
 	j2date(jd, &year, &month, &day);
 
 	CopyAppendFmt("%04d-%02d-%02d%s", (year <= 0) ? -year + 1 : year, month, day, (year <= 0) ? " BC" : "");
@@ -645,9 +816,9 @@ decode_date(const char *buffer, unsigned int buff_size, unsigned int *out_size)
 
 /* Decode a timestamp type */
 static int
-decode_timestamp(const char *buffer, unsigned int buff_size, unsigned int *out_size)
+decode_timestamp_internal(const char *buffer, unsigned int buff_size, unsigned int *out_size, bool with_timezone)
 {
-	const char *new_buffer = (const char *) TYPEALIGN(sizeof(int64), (uintptr_t) buffer);
+	const char *new_buffer = (const char *) LONGALIGN(buffer);
 	unsigned int delta = (unsigned int) ((uintptr_t) new_buffer - (uintptr_t) buffer);
 	int64		timestamp,
 				timestamp_sec;
@@ -668,6 +839,17 @@ decode_timestamp(const char *buffer, unsigned int buff_size, unsigned int *out_s
 	*out_size = sizeof(int64) + delta;
 	timestamp = *(int64 *) buffer;
 
+	if (timestamp == DT_NOBEGIN)
+	{
+		CopyAppend("-infinity");
+		return 0;
+	}
+	if (timestamp == DT_NOEND)
+	{
+		CopyAppend("infinity");
+		return 0;
+	}
+
 	jd = timestamp / USECS_PER_DAY;
 	if (jd != 0)
 		timestamp -= jd * USECS_PER_DAY;
@@ -684,20 +866,33 @@ decode_timestamp(const char *buffer, unsigned int buff_size, unsigned int *out_s
 	j2date(jd, &year, &month, &day);
 	timestamp_sec = timestamp / 1000000;
 
-	CopyAppendFmt("%04d-%02d-%02d %02" INT64_MODIFIER "d:%02" INT64_MODIFIER "d:%02" INT64_MODIFIER "d.%06" INT64_MODIFIER "d%s",
+	CopyAppendFmt("%04d-%02d-%02d %02" INT64_MODIFIER "d:%02" INT64_MODIFIER "d:%02" INT64_MODIFIER "d.%06" INT64_MODIFIER "d%s%s",
 				  (year <= 0) ? -year + 1 : year, month, day,
 				  timestamp_sec / 60 / 60, (timestamp_sec / 60) % 60, timestamp_sec % 60,
 				  timestamp % 1000000,
+				  with_timezone ? "+00" : "",
 				  (year <= 0) ? " BC" : "");
 
 	return 0;
 }
 
+static int
+decode_timestamp(const char *buffer, unsigned int buff_size, unsigned int *out_size)
+{
+	return decode_timestamp_internal(buffer, buff_size, out_size, false);
+}
+
+static int
+decode_timestamptz(const char *buffer, unsigned int buff_size, unsigned int *out_size)
+{
+	return decode_timestamp_internal(buffer, buff_size, out_size, true);
+}
+
 /* Decode a float4 type */
 static int
 decode_float4(const char *buffer, unsigned int buff_size, unsigned int *out_size)
 {
-	const char *new_buffer = (const char *) TYPEALIGN(sizeof(float), (uintptr_t) buffer);
+	const char *new_buffer = (const char *) INTALIGN(buffer);
 	unsigned int delta = (unsigned int) ((uintptr_t) new_buffer - (uintptr_t) buffer);
 
 	if (buff_size < delta)
@@ -718,7 +913,7 @@ decode_float4(const char *buffer, unsigned int buff_size, unsigned int *out_size
 static int
 decode_float8(const char *buffer, unsigned int buff_size, unsigned int *out_size)
 {
-	const char *new_buffer = (const char *) TYPEALIGN(sizeof(double), (uintptr_t) buffer);
+	const char *new_buffer = (const char *) DOUBLEALIGN(buffer);
 	unsigned int delta = (unsigned int) ((uintptr_t) new_buffer - (uintptr_t) buffer);
 
 	if (buff_size < delta)
@@ -758,7 +953,7 @@ static int
 decode_macaddr(const char *buffer, unsigned int buff_size, unsigned int *out_size)
 {
 	unsigned char macaddr[6];
-	const char *new_buffer = (const char *) TYPEALIGN(sizeof(int32), (uintptr_t) buffer);
+	const char *new_buffer = (const char *) INTALIGN(buffer);
 	unsigned int delta = (unsigned int) ((uintptr_t) new_buffer - (uintptr_t) buffer);
 
 	if (buff_size < delta)
@@ -794,23 +989,24 @@ decode_bool(const char *buffer, unsigned int buff_size, unsigned int *out_size)
 static int
 decode_name(const char *buffer, unsigned int buff_size, unsigned int *out_size)
 {
-	const char *new_buffer = (const char *) TYPEALIGN(sizeof(uint32), (uintptr_t) buffer);
-	unsigned int delta = (unsigned int) ((uintptr_t) new_buffer - (uintptr_t) buffer);
-
-	if (buff_size < delta)
+	if (buff_size < NAMEDATALEN)
 		return -1;
 
-	buff_size -= delta;
-	buffer = new_buffer;
-
-	if (buff_size < NAMEDATALEN)
-		return -2;
-
 	CopyAppendEncode(buffer, strnlen(buffer, NAMEDATALEN));
-	*out_size = NAMEDATALEN + delta;
+	*out_size = NAMEDATALEN;
 	return 0;
 }
 
+/*
+ * Decode numeric type.
+ */
+static int
+decode_numeric(const char *buffer, unsigned int buff_size, unsigned int *out_size)
+{
+       int result = extract_data(buffer, buff_size, out_size, &CopyAppendNumeric);
+       return result;
+}
+
 /* Decode a char type */
 static int
 decode_char(const char *buffer, unsigned int buff_size, unsigned int *out_size)
@@ -834,8 +1030,20 @@ decode_ignore(const char *buffer, unsigned int buff_size, unsigned int *out_size
 /* Decode char(N), varchar(N), text, json or xml types */
 static int
 decode_string(const char *buffer, unsigned int buff_size, unsigned int *out_size)
+{
+       int result = extract_data(buffer, buff_size, out_size, &CopyAppendEncode);
+       return result;
+}
+
+/*
+ * Align data, parse varlena header, detoast and decompress.
+ * Last parameters responds for actual parsing according to type.
+ */
+static int
+extract_data(const char *buffer, unsigned int buff_size, unsigned int *out_size, int (*parse_value)(const char *, int))
 {
 	int			padding = 0;
+	int			result	= 0;
 
 	/* Skip padding bytes. */
 	while (*buffer == 0x00)
@@ -854,14 +1062,13 @@ decode_string(const char *buffer, unsigned int buff_size, unsigned int *out_size
 		 * 00000001 1-byte length word, unaligned, TOAST pointer
 		 */
 		uint32		len = VARSIZE_EXTERNAL(buffer);
-		int			result = 0;
 
 		if (len > buff_size)
 			return -1;
 
 		if (blockOptions & BLOCK_DECODE_TOAST)
 		{
-			result = ReadStringFromToast(buffer, buff_size, out_size);
+			result = ReadStringFromToast(buffer, buff_size, out_size, parse_value);
 		}
 		else
 		{
@@ -883,9 +1090,9 @@ decode_string(const char *buffer, unsigned int buff_size, unsigned int *out_size
 		if (len > buff_size)
 			return -1;
 
-		CopyAppendEncode(buffer + 1, len - 1);
+		result = parse_value(buffer + 1, len - 1);
 		*out_size = padding + len;
-		return 0;
+		return result;
 	}
 
 	if (VARATT_IS_4B_U(buffer) && buff_size >= 4)
@@ -898,9 +1105,9 @@ decode_string(const char *buffer, unsigned int buff_size, unsigned int *out_size
 		if (len > buff_size)
 			return -1;
 
-		CopyAppendEncode(buffer + 4, len - 4);
+		result = parse_value(buffer + 4, len - 4);
 		*out_size = padding + len;
-		return 0;
+		return result;
 	}
 
 	if (VARATT_IS_4B_C(buffer) && buff_size >= 8)
@@ -911,7 +1118,9 @@ decode_string(const char *buffer, unsigned int buff_size, unsigned int *out_size
 		int						decompress_ret;
 		uint32					len = VARSIZE_4B(buffer);
 		uint32					decompressed_len = 0;
+#if PG_VERSION_NUM >= 140000
 		ToastCompressionId		cmid;
+#endif
 
 #if PG_VERSION_NUM >= 140000
 		decompressed_len = VARDATA_COMPRESSED_GET_EXTSIZE(buffer);
@@ -934,31 +1143,32 @@ decode_string(const char *buffer, unsigned int buff_size, unsigned int *out_size
 			return 0;
 		}
 
+#if PG_VERSION_NUM >= 140000
 		cmid = VARDATA_COMPRESSED_GET_COMPRESS_METHOD(buffer);
 		switch(cmid)
 		{
 			case TOAST_PGLZ_COMPRESSION_ID:
 				decompress_ret = pglz_decompress(VARDATA_4B_C(buffer), len - 2 * sizeof(uint32),
-												 decompress_tmp_buff, decompressed_len
-#if PG_VERSION_NUM >= 120000
-												 , true
-#endif
-												 );
+												 decompress_tmp_buff, decompressed_len, true);
 				break;
-			case TOAST_LZ4_COMPRESSION_ID:
 #ifdef USE_LZ4
+			case TOAST_LZ4_COMPRESSION_ID:
 				decompress_ret = LZ4_decompress_safe(VARDATA_4B_C(buffer), decompress_tmp_buff,
 													 len - 2 * sizeof(uint32), decompressed_len);
 				break;
-#else
-				printf("Error: compression method lz4 not supported.\n");
-				printf("Try to rebuild pg_filedump for PostgreSQL server of version 14+ with --with-lz4 option.\n");
-				return -2;
 #endif
 			default:
 				decompress_ret = -1;
 				break;
 		}
+#else /* PG_VERSION_NUM < 140000 */
+		decompress_ret = pglz_decompress(VARDATA_4B_C(buffer), len - 2 * sizeof(uint32),
+										 decompress_tmp_buff, decompressed_len
+#if PG_VERSION_NUM >= 120000
+										 , true
+#endif
+										 );
+#endif /* PG_VERSION_NUM >= 140000 */
 
 		if ((decompress_ret != decompressed_len) || (decompress_ret < 0))
 		{
@@ -968,9 +1178,9 @@ decode_string(const char *buffer, unsigned int buff_size, unsigned int *out_size
 			return 0;
 		}
 
-		CopyAppendEncode(decompress_tmp_buff, decompressed_len);
+		result = parse_value(decompress_tmp_buff, decompressed_len);
 		*out_size = padding + len;
-		return 0;
+		return result;
 	}
 
 	return -9;
@@ -1033,7 +1243,7 @@ FormatDecode(const char *tupleData, unsigned int tupleSize)
 	CopyFlush();
 }
 
-static int DumpCompressedString(const char *data, int32 compressed_size)
+static int DumpCompressedString(const char *data, int32 compressed_size, int (*parse_value)(const char *, int))
 {
 	int						decompress_ret;
 	char				   *decompress_tmp_buff = malloc(TOAST_COMPRESS_RAWSIZE(data));
@@ -1087,7 +1297,8 @@ static int DumpCompressedString(const char *data, int32 compressed_size)
 static int
 ReadStringFromToast(const char *buffer,
 		unsigned int buff_size,
-		unsigned int* out_size)
+		unsigned int* out_size,
+		int (*parse_value)(const char *, int))
 {
 	int		result = 0;
 
@@ -1160,9 +1371,9 @@ ReadStringFromToast(const char *buffer,
 			if (result == 0)
 			{
 				if (VARATT_EXTERNAL_IS_COMPRESSED(toast_ptr))
-					result = DumpCompressedString(toast_data, toast_ext_size);
+					result = DumpCompressedString(toast_data, toast_ext_size, parse_value);
 				else
-					CopyAppendEncode(toast_data, toast_ext_size);
+					result = parse_value(toast_data, toast_ext_size);
 			}
 			else
 			{
@@ -1191,8 +1402,7 @@ DecodeOidBinary(const char *buffer,
 		unsigned int *processed_size,
 		Oid *result)
 {
-	const char	   *new_buffer =
-		(const char*)TYPEALIGN(sizeof(Oid), (uintptr_t)buffer);
+	const char	   *new_buffer = (const char *) INTALIGN(buffer);
 	unsigned int	delta =
 		(unsigned int)((uintptr_t)new_buffer - (uintptr_t)buffer);
 
