@@ -3,7 +3,7 @@
  *				   formatting heap (data), index and control files.
  *
  * Copyright (c) 2002-2010 Red Hat, Inc.
- * Copyright (c) 2011-2021, PostgreSQL Global Development Group
+ * Copyright (c) 2011-2022, PostgreSQL Global Development Group
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -159,7 +159,7 @@ DisplayOptions(unsigned int validOptions)
 		printf
 			("\nVersion %s (for %s)"
 			 "\nCopyright (c) 2002-2010 Red Hat, Inc."
-		  "\nCopyright (c) 2011-2021, PostgreSQL Global Development Group\n",
+		  "\nCopyright (c) 2011-2022, PostgreSQL Global Development Group\n",
 			 FD_VERSION, FD_PG_VERSION);
 
 	printf
@@ -177,8 +177,8 @@ DisplayOptions(unsigned int validOptions)
 		 "  -D  Decode tuples using given comma separated list of types\n"
 		 "      Supported types:\n"
 		 "        bigint bigserial bool char charN date float float4 float8 int\n"
-		 "        json macaddr name oid real serial smallint smallserial text\n"
-		 "        time timestamp timetz uuid varchar varcharN xid xml\n"
+		 "        json macaddr name numeric oid real serial smallint smallserial text\n"
+		 "        time timestamp timestamptz timetz uuid varchar varcharN xid xml\n"
 		 "      ~ ignores all attributes left in a tuple\n"
 		 "  -f  Display formatted block content dump along with interpretation\n"
 		 "  -h  Display this information\n"
